/* MY_TCO vB.1 (TCO XpressNet + I2C)

   Copyright (c) Salustiano Moreno, Summer'2018  All right reserved.

   XpressNet library, Copyright (c) Philipp Gahtow  All right reserved. http://pgahtow.de/wiki/index.php?title=XpressNet

This software and associated files is a DIY project not intended for commercial use.

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

This software uses libraries with different licenses, follow all their different terms included.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ALOW FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


DESCRIPCION GENERAL =================================================================================================================================================

Este sketch utiliza la librería XpressNet from https://sourceforge.net/projects/pgahtow/files/Arduino%20%28v1.0%29%20libaries/XpressNet.zip/download
Usando funciones de esta librería, ee pretende construir un TCO XpressNet, usando el bus se envian comandos y que la central gestione los accesorios DCC, see usan
teclas como elemento de entrada y LED's como indicadores de salida. 

Algo que se ha tenido en cuenta para desarrollar en este sketch, es dedicar el mayor tiempo de proceso a "escuchar" el bus XpressNet y no perder ningun comando a tratar
en el TCO, se ha intentado optimizar los procesos evitando las busquedas secuenciales (pooling) de los puertos E/S, indexando matrices para acceder directamante a la 
posición donde está el dato requerido, preprocesando calculos en el arranque (setup) para evitar hacerlos posteriormente de forma repetitiva, etc.

Para poder disponer de teclas y LEDs suficientes, sin depender del número de E/S del Arduino usado, se usan expansores E/S PCF8574/A vía el bus I2C.
El PCF8574 es un circuito integrado fácil de usar y encontrar, además no es muy caro, nos ofrece 8 puertos E/S y una línea se interrupción que se activa al 
detectar un cambio en un puerto definido como entrada, según el direccionamiento de este circuito y usando laas dos versiones PCF8574 y PCF8574A, podemos conectar 
hasta 16, con esto tenemos 16*8 E/S, son 128 en total.

En cada puerto podemos conectar un LED (salida) o una TECLA (entrada), un puerto actuará como entrada o salida dependiendo de la configuración 
de la matriz PortManaged[].

Todos los puertos han de estar configurados, aunque no tengan conexión alguna.

Para evitar la lectura continua de los puertos de entrada buscando una pulsación, se usa una interrupción proporcionada por el PCF al cambiar el estado de alguuno
de sus puertos, de esta forma solo buscamos el puerto correspondiente, cuando efectivamente se pulsa una tecla.

Cuando se detecta una pulsación, usamos matrices como indice para acceder directamente a laposición en la matris donde están los parámetros de cada accesorio, 
sin tener que buscar su dirección DCC a lo largo de una lista.

A cada puerto de entrada (tecla), se le puede asignar una "acción", normalmente ejecutar una ruta.
Una ruta define una secuencia de ordenes que se enviará a la central para cambiar el estado accesorios, esta secuencia puede contener uno o varios accesorios, 
hay otras acciones predefinidas (ver mas adelante).

Direcciones y puertos virtuales:
Se pueden definir direcciones de accesorio virtuales para ejecutar rutas virtuales, estas direcciones deben de estar libres en el trazado.
A una dirección virtual le podemos asignar una ruta virtual que se ejecutará cuando se reciba un cambio de estado de esta, de esta forma podemos 
ejecutar una ruta desde el mando de la central con solo una pulsación.

Hay una dirección virtual especial que se define en la variable DIRBLOQUEO, según el estado de esta dirección se bloqueará o desbloqueará el teclado del TCO, con 
esto podemos inhabilitar el TCO si se está controlando el trazado desde un PC y así evitar contra ordenes a la central entre PC y TCO, los puertos configurados
como salida (LED's) siguen funcionando normalmente.

DESCRIPCION DE LOS PROCESOS BASICOS =================================================================================================================================
 
- 1 >> Escucha del bus XpressNet:
La función "XpressNet.receive()" se ejecuta de forma continua en el bucle principal "loop".
Al detectar un comando en el bus, referente a un cambio en un accesorio, se ejecuta la función "notifyTrnt()", que nos proporciona el número del accesorio y su estado
En esta función identificamos si el accesorio forma parte de los gestionados por el TCO y lanzamos la actualización del panel de LED's con la función "UpdateLED()".

- 2 >> Cambio de estado de un accesorio desde el TCO
Al pulsar una tecla se activa la interrupción de los PCF's, detectamos el bloque y bit activado (# de puerto) con la función "KeyPressed()"
Con el # de puerto accedemos a los datos de la "acción" a ejecutar en la matriz "PortManaged[]", si es una ruta encontramos el tipo de ruta y la secuencia de 
accesorios, según el tipo de acción en la estructura "case" del bucle principal lanzaremos la función correspondiente, ej."launchRoute_1()" o "launchRoute_2()"
Con estas funciones, usaremos la matriz "AccIndex[]" que servirá de indice para acceder directamente a los datos del accesorio definidos en la matriz "AccManaged[]"
Con la función "updateAcc()" enviamos los comandos a la central para el cambio de estado usando la función "XpressNet.setTrntPos()", al cambisr el estade del 
accesorio, la función "XpressNet.receive()" detecta el cambio y se actualiza el panel según el proceso antes definido "Escucha del bus Xpressnet"
Al inicio del proceso se deshabilitan las interrupciones para no atender a una nueva pulsación hasta finalizar 

- 3 >> Petición del estado de un accesorio
Pedimos el estado de cada accesorio de la matrix "AccManaged[]", ejecutando la función "XpressNet.getTrntInfo()", esto provoca que la central publique el estado 
del accesorio, la función "XpressNet.receive()" detecta el cambio y se actualiza el panel según el proceso "Escucha del bus Xpressnet"

- 4 >> Arranque y control de estado de la central
En el bucle principal "loop()" llamamos a la función "XpressNet.receive()", si hay cambio en el estado de la central se ejecuta la función "notifyXNetPower" que inicializa el estado de la central
si el valor devuelto != 0, esperamos en el bucle "loop" hasta que la central esté disponible, la acción "tipo=5", define la tecla para el STOP de emergencia, bloquea 
y desbloquea la central, al inicio si la central está deshabilitada la primara activación se ha de hacer desde el mando no desde el TCO
Cada vez que la central se sale de un STOP, se pide el estado de los accesorios según el proceso "Petición del estado de un accesorio"

AYUDA AL DEBUG =======================================================================================================================================================
Al usar un Arduino UNO o NANO solo tenemos un puerto serie disponible, conectando un LCD I2C al bus podemos usarlo para mostrar el estado de variables a lo largo 
del sketch con la función DebugLCD()
Hay partes del sketch que se pueden ejecutar sin ocupar el puerto serie leyendo el bus XpressNet, entonces podemos habilitar el terminal serie para mostrar el 
estado de variables a lo largo del sketch, la función updateAcc() está duplicada para que la actualización de los LED's funcione
Cuando se trabaje sin conexión al bus Xpressnet, se puede desconectar el MAX485, de esa forma cargaremos el sketch sin desconectar los pins RX/TX

VELOCIDAD DEL BUS I2C (USAR CON PRECAUCION) ==========================================================================================================================
La valocidad estandar del bus I2c es de 100KHz (CLOCKI2C=0), con la función "setclock()" la fijamos 400KHz (CLOCKI2C=1)
Según sea el cableado de las líneas SDA, SCL y GND esto puede causar problemas en la lectura/escritura de datos, en ese caso fijarla a 100KHz
*/

#define DEBUG_LCD 0                         // 1 para activar el debug vía LCD
#define DEBUG_SERIE 0                       // 1 para activar la comunicación serie

#if DEBUG_SERIE == 0
  #include <XpressNet.h>                    // Librería que gestiona la comunicación con la central vía el bus XpressNet
  XpressNetClass XpressNet;                 // Inicializamos XpressNet
  #define XNetAddress 30                    // Dirección XpressNet
  #define XNetSRPin 7                       // Conexión Max485 Recepción/Envío
#endif
#if DEBUG_LCD == 1                          // Inicializamos el LCD en la dirección 0x27, con 16 caracteres y 2 líneas
  #include <LiquidCrystal_I2C.h>  
  LiquidCrystal_I2C lcd(0x27,16,2);
#endif

#include <Wire.h>                           // Librería I2C para el control de los PCF's y el LCD si es necesario
#include <avr/pgmspace.h>                   // Librería que gestiona la memoria, para almacenar datos del proceso en memoria de programa y liberar memoria para variables

// DEFINICIONES DE CONSTANTES ==========================================================================================================================================
#define CLOCKI2C 1                          // Velocidad I2C 0/1 = 100KHz/400KHz
#define LEDFLASH 100                        // Frecuencia de parpadeo de un LED con millis()
#define LED_CENTRAL 5                       // Salida para mostrar el estado de la central
#define LED_CENTRAL_ONOFF {prevMillis = millis(); flagMillis=!flagMillis; digitalWrite(LED_CENTRAL, flagMillis);} // Comandos control LED central
#define LED_CENTRAL_ON digitalWrite(LED_CENTRAL, HIGH)     // Salida para mostrar estado de la central si normal
#define LED_BLOCKED 4                       // Salida para mostrar el bloqueo del TCO, usamos un LED conectado a un pin del Arduino
#define LED_BLOCKED_ONOFF digitalWrite(LED_BLOCKED, TCOblocked) // Control LED bloqueo del TCO
#define LED_XNET 12                         // Indicación gestión de órdenes recibidas por el bus
#define DELAY_RECEIVE 1                     // Espera (ms) para cargardatos tras ejecutar XpressReceive
#define DELAY_ASKACC 100                    // Espera (ms * DELAY_RECEIVE) para pedir el estado de accesorios
#define DELAY_ACC 50                        // Retardo ordenes entre accesorios, 0-255ms
#define DELAY_PULSE 50                      // Duración de la orden de cambio de estada para un accesorio, 0-255ms
#define NUM_BLOCKS 3                        // Número de bloques (#PCF's) conectados al bus I2C para la gestion de puertos E/S
#define NUM_PORT_MANAGED NUM_BLOCKS * 8     // Número de puertos a gestionar teclas/rutas y LEDs
#define NUM_ACC_MANAGED 8                   // Número de accesorios reales a manejar
#define VIR_ACC_MANAGED 0 + 1               // Número direcciones virtuales manejadas mas 1, la dirección de bloqueo
#define VIR_PORT_MANAGED 0                  // Número de puertos virtuales
#define MAX_ACC_ROUTE 3                     // Máximo número de accesorios en una ruta, se usa para dimensionar la matriz PortManaged
#define MIN_ACC_NUM 3                       // La dirección DCC usada mas baja para un accesorio real
#define MAX_ACC_REAL 11                     // La dirección DCC mas alta usada para un accesorio real
#define ADD_BLOCKED 12                      // Definimos la dirección DCC para el bloqueo del TCO, siempre la más alta de las virtuales
#define MAX_ACC_NUM ADD_BLOCKED             // La dirección DCC usada mas alta, incluyendo las virtuales si existen, será la dirección de bloqueo del TCO
#define PIN_INTERRUPT 3                     // Pin interrupción para lectura teclas
#define STOP_KEY 99                         // Acción a asignar a un port como tecla de STOP de emergencia
#define INT_ON {attachInterrupt(digitalPinToInterrupt(PIN_INTERRUPT), ServiceINT, FALLING); digitalWrite(LED_XNET, HIGH);}// Activamos el servicio de interrupción en el flanco de bajada
#define INT_OFF {detachInterrupt(digitalPinToInterrupt(PIN_INTERRUPT)); digitalWrite(LED_XNET, LOW);} // Deshabilitamos interrupción

// INICIALIZACION DE VARIABLES ===============================================================================================================================================================
//long cur_Millis = 0;                      // Variable para almacenar valor actual de millis(), de momento no se usa
long prevMillis = 0;                        // Variable para almacenar valor previo de millis()
boolean flagMillis = false;                 // Variable para uso en contadores
byte n = 0;                                 // Variable para uso en contadores
boolean startup = true;                     // Control startup
boolean TCOblocked = true;                  // Control bloqueo TCO, false = no bloqueado, bloqueado al inicio
boolean interruptON = true;                 // Control interrupción
byte centralStatus = 255;                   // Control de estado de la central, por defecto inactiva al inicio
byte keyInArray = 0;                        // Posición en el array correspondiente a la tecla pulsada
const byte blockAdd[NUM_BLOCKS] = {0x20,0x21,0x23};  // Direcciones de los PCF's (bloques)
byte blockStatus[NUM_BLOCKS] = {};          // Matriz para guardar los estados de los puertos
byte blockMask[NUM_BLOCKS] = {};            // Matriz para guarda guarda una mascara de cada bloque para filtrar lectura/escritura de teclas/LED's


/* Matriz de accesorios a controlar ---------------------------------------------------------------------------------------------------------------------------------------------------------
En la matriz NUM_ACC_MANAGED, se definen los parámetros de los accesorios a gestionar, no hay límite en en número de accesorios en la matriz
La gestión de accesorios según la tecla pulsade, se configura en la matriz NUM_PORT_MANAGED (puertos)  */
#define ACC_NUM 0  // Columna 0: Número de accesorio (dirección DCC), dato obligatorío
#define ACC_LED 1  // Columna 1: Port(+1) I2C correspondiente al LED del accesorio, 0=no LED, dato obligatorío
#define ACC_INV 2  // Columna 1: Invertir orden al mover el accesorio
#define ACC_HIG 3  // Columna 2: En el Setup guardamos el byte alto de cada accesorio, se calcula en el setup
#define ACC_LOW 4  // Columna 3: En el Setup guardamos el byte bajo de cada accesorio, se calcula en el setup  
#define ACC_STA 5  // Columna 4: Estado del accesorio, se actualizará mediante funciones de la liberería XpressNet 
int AccManaged[NUM_ACC_MANAGED + 1 + VIR_ACC_MANAGED][5] = {  
{},     // Despreciamos la fila 0 para que los contadores sean a partir de la #1
{3,21,0},  
{4,24,0},
{5,22,0},
{6,23,0},
{8,11,1},
{9,13,0},
{10,16,1},
{11,8,0},
// Direcciones virtuales,han de ser direcciones no usadas por ningún accesorio en todo el trazado

//{12,17},           // cuando se reciba una orden para la dirección DCC #12 se lanzará la ruta #17
//{13,18},
//{14,19},
{ADD_BLOCKED,0}      // incluimos la direccion de blqueo del TCO dentro del grupo de las virtuales
};

/* Creamos una matriz vacia (AccIndex[]), para indexar la matriz AccManaged y acceder directamente a la fila correspondiente al # de accesorio requerido
La dimension de esta matriz será el rango entre la dirección de accesorio más baja (MIN_ACC_NUM) y la mas alta (MAX_ACC_NUM) de los accesorios definidos,
Si este rango el muy grande podemos generar problemas por falta de de memoria SRAM
Por ejemplo si solo usamos dos accesorios teniendo en cuenta que consumimos dos bytes por accesorio:
- Si la direccion del primero es la #1 y la del segundo es la #1000, la matriz tendrá 1000 filas y ocupará 1000*2 = 2000bytes (~2Kb)
- Si la direccion del primero es la #1 y la del segundo es la #2, la matriz tendrá 2 filas y ocupará 2*2 = 4bytes 

Tenemos en cuenta el número de accesorio mas bajo (MIN_ACC_NUM), para situarlo en la posición 1 de la matriz, por ejemplo, si el # de accesorio mas bajo es el 301 y
el más alto el 320, la matriz (AccIndex) tendrá solamante 20 posiciones.

En el setup, inicializamos la matriz AccIndex con la fila que ocupa cada accesorio en la matriz AccManaged, para usarla como indice y tener acceso directo 
a los parámetros de cada accesorio sin tener que buscar uno a uno en las filas de la matriz
*/
byte AccIndex[MAX_ACC_NUM - MIN_ACC_NUM + 2 + VIR_ACC_MANAGED];

/* Gestion de los puertos de disponibles -----------------------------------------------------------------------------------------------------------------------------------------------------
Tenemos tantos puertos como PCF's utilizados * 8, un puerto puede ser de lectuta (TECLA) o de salida (LED)
Hay que definir los parametros necesarios en todos los puertos en la matriz PortManaged[], aunque no tenga nada conectado
En la columna 0 definimos el tipo de acción a ejecutar, por ejemplo:
0 = control de indicadores LED
Otro valor determina la acción a ejecutar cuando se reciba una pusación de una TECLA, posibles valores:
1 = ejecutar ruta tipo 1
2 = ejecutar ruta tipo 2
3 = reset TCO a posiciones iniciales
4 = ejecutar varias rutas conjuntamente
5 = control estado de la centrar on/off
* Parametros para configuración como salida de un puerto para gestionar un LED
Columna 0: 0 configuración de salida, para gestionar un LED en ese puerto
Columna 1: 1 para invertir estado LED con respecto al estado del accesorio, 0 para no invertir  (NO USADO EN ESTA VERSION)
* Parametros de un puerto como entrada, para leer una TECLA
Podemos asignar un accesorio o varios a cada puerto, en una ruta puede haber uno o más accesorios definidoscon
Se pueden gestionar dos tipos de ruta, 1/2
Columna 0: 1, 2 según tipo de ruta 
Ruta tipo 1: se invierte el estado de los accesorios definidos, se usa una columna por accesorio, #fila en AccManaged
Ruta tipo 2: se define el estado a enviar al accesorio, se usan un par de columnas para cada accesorio, #fila en AccManaged y estado 1=ON (+)
Esto permite varias posibilidades para controlar accesorios:
- Invertir el estado de un/unos accesorio/s usando una tecla
- Cambiar el estado de un/unos accesorio/s, a un estado definido con una tecla (usando dos teclas por accesorio) 
- Cambiar el estado de varios accesorios (ruta) con una tecla, por ejemplo desvíos y semaforos
* Ültima columna: siempre valor 0 <---- IMPORTANTE
* Resto de columnas si tipo ruta=1, cada columna corresponderá a la fila donde están los parametros del accesorio en la matriz AccManaged
* Resto de columnas si tipo ruta=2, por cada par de columnas, la 1ª corresponderá a la fila donde están los parametros del accesorio en la matriz AccManaged, 
      la 2ª corresposde al estado del accesorio 1=ON (+)    
Ejemplos de configuración de puertos:
{0} = tipo 0, salida para el control de un LED
[1,1,2,3,0} = tipo 1, entrada (tecla), para cada columna, se invertirá es estado de los accesorios definidos en las 1,2,3 e la matrix AccManaged[], siempre 0 al final
{2,1,0,2,1} = tipo 3, entrada {tecla), para cada par de columnas, al accesorio definido en la 1ª se le asignará el estado definido en la 2ª, siempre 0 al final 
{3} = tipo 3, entrada {tecla),se ejecuta la ruta definida en la matriz RutaReset[]
{4,3,12,0} = tipo 4, entrada {tecla), se ejecuta cada una de las rutas definidas en cada columna, siempre 0 al final 
{STOP_KEY} = entrada {tecla), control estado de la centrar on/off
Con todo esto podemos establecer varías formas de funcionamiento:
- Configurar dos teclas para mover un desvío recto/desviado
- Configurar una tecla para mover un desvío alternando recto/desviado
- Configurar rutas que posiciones desvíos en una posición concreta
- Controlar otros accesorios en una ruta, como semaforos, reles, etc.
- Lanzar las rutas delde el mando de la central configurando accesorios virtuales
- Programar otras funciones (como la 3) modificando la estructura "case" en la función loop
*/
// Utilizamos PROGMEM para guardar la matriz PortManaged en la memoria de programa y no ocupar memoria para las variables (SRAM)
const PROGMEM byte PortManaged[NUM_PORT_MANAGED + VIR_PORT_MANAGED + 1][(MAX_ACC_ROUTE * 2) + 2] = {  // Tipo de ruta, #fila/s en matriz accesorios, última posición siempre 0 
{}, //Despreciamos la posición 0 para asignar cada fila una tecla contando a partir de la #1

// Parámetros de los puertos 1 a 8, de bloque definido en blockAdd[0]
{STOP_KEY,0}, // STOP_KEY
{1,8},  // Tecla no usada actualmente
{0},
{0},
{0},
{0},
{0},
{0},  //------------ 8

// Parámetros de los puertos 9 a 16, de bloque definido en blockAdd[1]
{1,7,0}, // La columna 0 define el tipo de ruta, tipo = 1 para mover accesorios invirtiendo la posición actual, última posición siempre 0 
{1,5,0}, // Puede haber uno a más accesorios en cada ruta
{0},
{3,0},
{0},
{1,6,0},
{0},    // 
{0},    // ON/OFF central

// Parámetros de los puertos 17 a 24, de bloque definido en blockAdd[0]
{1,1,0},
{1,2,0},
{1,4,0},
{1,3,0},
{0},
{0},
{0},
{0}, 

// Después de definir los puertos según los PCF instalados, definimos los puertos virtuales
//{2,1,1,2,1,3,1,0},
//{2,1,0,2,0,3,0,0},
//{1,4,0},
};


// Definimos una ruta (tipo 2) para la podemos usar para llamarla desde una tecla (tipo = 3)
// Utilizamos PROGMEM para guardar la matriz RutaReset en la memoria de programa y no ocupar memoria para las variables (SRAM)
const PROGMEM byte RutaReset[]={2,1,0,2,0,3,0,4,0,7,0,5,0,6,0,8,0,0}; // Ruta que define el estado reset de los accesorios
 
// INICIALIZAMOS ENTORNO ==============================================================================================================================================================================
void setup() {
  pinMode(LED_BLOCKED, OUTPUT);    // Inicializamos LED bloqueo
  pinMode(LED_CENTRAL, OUTPUT);    // Inicializamos LED central
  LED_BLOCKED_ONOFF;               // Mostramos estado de TCO bloqueado al startup
  // Inicializamos LCD
  #if DEBUG_LCD == 1
     lcd.begin();                      
     lcd.backlight();
     DebugLCD("L0: ",0,"L1: ",0); delay(1);
  #endif 
  // Inicializamos comunicación serie
  #if DEBUG_SERIE == 1
     Serial.begin(9600);
     Serial.println("Comunicación serie activa");
  #endif
  // Inicializamos I2C 
  Wire.begin();
  #if CLOCKI2C == 1
    Wire.setClock(400000L);                                      // Fijamos velocidad del bus I2C a 400KHz           
  #endif
  byte n = 1;                                                    // Variable para contador
  for ( n = 1; n <= NUM_ACC_MANAGED; n++) {                      // Guardamos bytes alto y bajo de la dirección DCC de cada accesorio
    AccManaged[n][ACC_HIG] = highByte(AccManaged[n][0]);         // Byte alto según #acc
    AccManaged[n][ACC_LOW] = lowByte(AccManaged[n][ACC_NUM]+1);  // Byte bajo según #acc                
    AccIndex[AccManaged[n][ACC_NUM] - MIN_ACC_NUM + 1] = n;      // Guardamos el número de fila del accesorio en la matriz AccManaged  
  }
  for (n = 0; n< NUM_BLOCKS; n++){                               // Generamos una mascara para cada bloque, pera filtrar las lecturas/escrituras a teclas/LED's
    byte mascara = 0;
    for (byte i = 8; i!=0; i--){
      if (pgm_read_byte(&PortManaged[(n*8) + i][0])==0) {mascara=(mascara << 1)| 0;}
      else{mascara=(mascara << 1)| 1;}
    }
    blockMask[n] = mascara;
  }
  // Inicializamos XpressNet  
  #if DEBUG_SERIE == 0
    XpressNet.start(XNetAddress, XNetSRPin); 
    delay(50);
  #endif
  // Inicializamos proceso para manejar la interrupción que controla si hay pulsación, la interrupción ejecuta la función ServiceINT
  pinMode(PIN_INTERRUPT, INPUT_PULLUP);
  pinMode(LED_XNET, OUTPUT);       // Inicializamos LED indicador de órdenes recibidas por el bus
  digitalWrite(LED_XNET, LOW);  
  INT_ON;                          // Activamos el servicio de interrupción en el flanco de bajada
  interruptON = false;
  TCOblocked = false;              // Habilitamos TCO
  LED_BLOCKED_ONOFF;               // Mostramos estado de bloqueo
}

// Visualización en LCD -----------------------------------------------------------------------------------------------------------------------------------------------
#if DEBUG_LCD == 1
void DebugLCD(String Texto_1, int Filaa_1, String Texto_2,int Filaa_2){ // Texto línea 1, variable línea 1, texto línea 2, variable línea 2
   String Filaa_0="                ";
   lcd.setCursor(0, 0);
   lcd.print(Texto_1+Filaa_1+Filaa_0);
   lcd.setCursor(0, 1);
   lcd.print(Texto_2+Filaa_2+Filaa_0);  
}
#endif

// Proceso de rutas, tipo 1 -------------------------------------------------------------------------------------------------------------------------------------------
void launchRoute_1(byte NumRuta){
  byte Contador = 1;
  while(pgm_read_byte(&PortManaged[NumRuta][Contador]) != 0){                            // Procesamos las columnas de la matriz RutasManaged
    AccManaged[pgm_read_byte(&PortManaged[NumRuta][Contador])][ACC_STA] = !AccManaged[pgm_read_byte(&PortManaged[NumRuta][Contador])][ACC_STA];  // Invertimos estado de accesorio
    updateAcc(pgm_read_byte(&PortManaged[NumRuta][Contador]));                           // Actualizamos el estado del accesorio
    delay(DELAY_ACC);                                                                    // Intercalamos retardo entre ordenes DCC
    Contador++;
  }
}

// Proceso de rutas, tipo 2 -------------------------------------------------------------------------------------------------------------------------------------------
void launchRoute_2(byte NumRuta){
  byte Contador = 1;
  while(pgm_read_byte(&PortManaged[NumRuta][Contador]) != 0){                            // Procesamos las columnas de la matriz RutasManaged
    AccManaged[pgm_read_byte(&PortManaged[NumRuta][Contador])][ACC_STA] = pgm_read_byte(&PortManaged[NumRuta][Contador + 1]); // Asignamos el estado definido
    updateAcc(pgm_read_byte(&PortManaged[NumRuta][Contador]));                           // Actualizamos el estado del accesorio
    delay(DELAY_ACC);                                                                    // Intercalamos retardo entre ordenes DCC
    Contador = Contador + 2;
  }
}

// Procesaos no definida en array -------------------------------------------------------------------------------------------------------------------------------------
void launchRoute(byte RutaName[]){
  byte n = 1;
  while(pgm_read_byte(&RutaName[n]) != 0){                                               // Procesamos las columnas de la matriz RutasManaged
    AccManaged[pgm_read_byte(&RutaName[n])][ACC_STA] = pgm_read_byte(&RutaName[n + 1]);  // Asignamos el estado del accesorio definido 
    updateAcc(pgm_read_byte(&RutaName[n]));                                              // Actualizamos el estado del accesorio
    delay(DELAY_ACC);                                                                    // Intercalamos retardo entre accesorios
    n += 2;
  }
}

// Procesaos RutaReset[] -------------------------------------------------------------------------------------------------------------------------------------
void launchReset(){
  byte n = 1;
  while(pgm_read_byte(&RutaReset[n]) != 0){                                               // Procesamos las columnas de la matriz RutasManaged
    AccManaged[pgm_read_byte(&RutaReset[n])][ACC_STA] = pgm_read_byte(&RutaReset[n + 1]);  // Asignamos el estado del accesorio definido 
    updateAcc(pgm_read_byte(&RutaReset[n]));                                              // Actualizamos el estado del accesorio
    delay(DELAY_ACC);                                                                     // Intercalamos retardo entre accesorios
    n += 2;
  }
}

// Calculamos # de la tecla pulsada ------------------------------------------------------------------------------------------------------------------------------
byte KeyPressed() {
  byte Block = 0;
  byte BitNum = 0;
  byte n = 0;
  do {
    BitNum = 0;    
    Wire.requestFrom(blockAdd[Block], 1); // Leemos 1 byte de cada bloque, hasta encontrar un valor != a 0, luego buscaremos el bit correspondiente a la tecla pulsada
    while (Wire.available())
    n = ~Wire.read() & blockMask[Block];  // Leemos haciendo NOT para tener solo un 1 en el bit que marca la posicion, obtenemos los bit que corresponden a TECLAS 
    //n = ~Wire.read();
    while (n != 0) {                      // Si el valor leido !=0, buscamos el #bloque y #bit que corresponde a la tecla pulsada
      BitNum++;
      n = n >> 1;
    }
    Block++;
  } while (Block < NUM_BLOCKS && BitNum == 0);
  if (BitNum == 0)return 0;
  else return((Block - 1) * 8) + BitNum;  // Número de tecla pulsada según el bloque y bit 
}

// Deshabilidamos servicio de interrupcion para no aceptar otra pulsación ----------------------------------------------------------------------------------------
void ServiceINT(){
  INT_OFF;  // Deshabilitamos interrupción
  interruptON = true;
}

// Actualizamos salidas (LED's) ----------------------------------------------------------------------------------------------------------------------------------
void UpdateLED(byte PosLED, boolean StatusLED) { 
  byte BlockActual = (PosLED - 1) / 8;                                  // Calculo del bloque de LED
  byte BitActual = PosLED - 1 - (BlockActual * 8);                      // Calculo bit a cambiar
/*if(pgm_read_byte(&PortManaged[PosLED][1])){  // Si el LED ha de actuar invertido al estado del accesorio (NO USADO EN ESTA VERSION) 
    StatusLED = !StatusLED;                    
    pgm_read_byte(&PortManaged[PosLED][1]) = !pgm_read_byte(&PortManaged[PosLED][1]);
}*/
  bitWrite(blockStatus[BlockActual], BitActual, StatusLED);             // Actualizamos el bit corrrspondiente a la salida del PCF donde está el LED conectado
  // Actualizamos el puerto correspondiente al LED en su bloque, escribimos vía I2C
  blockStatus[BlockActual] = blockStatus[BlockActual] | blockMask[BlockActual];
  Wire.beginTransmission(blockAdd[BlockActual]);
  Wire.write(blockStatus[BlockActual]);
  Wire.endTransmission();
}

// Funciones que se compilan en según las opciones de debug ==========================================================================================================
#if DEBUG_SERIE == 1  // este bloque se compila si se habilita la comunicación serie

// Actualiza estado de accsorio si está activo la comunicación seris -------------------------------------------------------------------------------------------------
void updateAcc(byte keyInArray) {
  Serial.print(keyInArray);Serial.print(" - ");
  Serial.print((AccManaged[keyInArray][ACC_NUM]));  
  if (AccManaged[keyInArray][ACC_STA]; == 1){ 
    AccManaged[keyInArray][ACC_STA]=1;
    Serial.println("  switch RECTO");   
  }
  else {
    AccManaged[keyInArray][ACC_STA]=0;
    Serial.println("  switch CURVA");  
  }
  UpdateLED(AccManaged[keyInArray][ACC_LED], AccManaged[keyInArray][ACC_STA]);
}
#endif

//====================================================================================================================================================================
#if DEBUG_SERIE == 0 // este bloque se compila si NO se habilita la comunicaciín serie, entonces se definen/usan los accesos a funciones de la librería XpressNet 

// Se actualiza el estado de un accesorio mediante la función "setTrntPos" -------------------------------------------------------------------------------------------
void updateAcc(byte keyInArray) {
  if (AccManaged[keyInArray][ACC_STA] == 1){ 
    XpressNet.setTrntPos(AccManaged[keyInArray][ACC_HIG], AccManaged[keyInArray][ACC_LOW] - 2, B1000); //switch RECTO   
    delay(DELAY_PULSE);
    XpressNet.setTrntPos(AccManaged[keyInArray][ACC_HIG], AccManaged[keyInArray][ACC_LOW] - 2, B0000); //switch RECTO       
  }
  else {
    XpressNet.setTrntPos(AccManaged[keyInArray][ACC_HIG], AccManaged[keyInArray][ACC_LOW] - 2, B1001); //switch DESVIADO  
    delay(DELAY_PULSE);
    XpressNet.setTrntPos(AccManaged[keyInArray][ACC_HIG], AccManaged[keyInArray][ACC_LOW] - 2, B0001); //switch DESVIADO  
  }
}

// Notificación de cambio en un accesorio, esta función se lanza desde la libería XpressNet ----------------------------------------------------------------------
void notifyTrnt(uint8_t Adr_High, uint8_t Adr_Low, uint8_t Pos) {
  if (keyInArray == 0) INT_OFF;                               // Si la orden de cambio del accesorio no proviene de una tecla pulsada se deshabilita la interrupción para no leer teclas  
  int acc_Num = (Adr_High * 256) + Adr_Low + 1;               // Calculamos dirección DCC 
  if(acc_Num > MAX_ACC_NUM || acc_Num < MIN_ACC_NUM) {
    if (keyInArray == 0) INT_ON;                              // Se habilita la lectura d eteclas
    return;  // despreciamos si la direción DCC, está fuera del rango de los accesorios usados                  
  }
  byte FilaAcc = acc_Num - MIN_ACC_NUM + 1;                   // Calculamos la fila en matriz AccIndex  
  if (acc_Num > MAX_ACC_REAL){                                // verificamos si se trata de un accesorio virtual
    if (acc_Num == ADD_BLOCKED){                              // Controlamos petición de bloqueo del TCO
      if(Pos == B01) TCOblocked = true;                       // Si Pos = 1, activamos bloqueo
      else TCOblocked = false;                                // Si Pos = 0, desactivamos bloqueo
      LED_BLOCKED_ONOFF;                                      // Mostramos estado de bloqueo
      if (keyInArray == 0) INT_ON;                            // Se habilita la lectura d eteclas
      return;              
    }
    if (pgm_read_byte(&PortManaged[AccManaged[FilaAcc-1][ACC_LED]][0]) == 1){   // determinamos acción en función del tipo de ruta    
      launchRoute_1(AccManaged[FilaAcc-1][ACC_LED]); }
    else {launchRoute_2(AccManaged[FilaAcc-1][ACC_LED]);}
    if (keyInArray == 0) INT_ON;                              // Se habilita la lectura d eteclas
    return;
  }  
  if(AccIndex[FilaAcc]!=0 ) {                                 // despreciamos #acc no usados, si es un #acc de los usados, actualizamos su estado
    #if DEBUG_LCD == 1
      DebugLCD("Acc: ",acc_Num,"Pos: ",Pos);
    #endif
    if (AccManaged[keyInArray][ACC_INV] == 1){                // Invertimos operación del accesorio
      if (Pos == B10) AccManaged[AccIndex[FilaAcc] ][ACC_STA] = 1; // ----
      if (Pos == B01) AccManaged[AccIndex[FilaAcc] ][ACC_STA] = 0; // ++++
    }
    else{                                                     // Operación normal sin invertir operación
      if (Pos == B10) AccManaged[AccIndex[FilaAcc] ][ACC_STA] = 0; // ----
      if (Pos == B01) AccManaged[AccIndex[FilaAcc] ][ACC_STA] = 1; // ++++
    }
    if (AccManaged[AccIndex[FilaAcc]][ACC_LED]){              // Actualizamos LED, si != 0 
      UpdateLED(AccManaged[AccIndex[FilaAcc]][ACC_LED],AccManaged[AccIndex[FilaAcc]][ACC_STA]);  
    }
  }
  if (keyInArray == 0) INT_ON;                                // Se habilita la lectura d eteclas
}

// Preguntamos por el estado de los accesorios ------------------------------------------------------------------------------------------------------------------
void AskAccPos(){ 
  for (byte n=1; n<=NUM_ACC_MANAGED;n++) {
    XpressNet.getTrntInfo(AccManaged[n][ACC_HIG], AccManaged[n][ACC_LOW] -2);   //@2 pedimos la posición de un accesorio con un intervalo DELAY_ASKACC * DELAY_RECEIVE    
    for (int i=0; i<DELAY_ASKACC; i++) {                                        // vamos recibiendo datos para que trabaje la librería
      XpressNet.receive();
      delay (DELAY_RECEIVE);
    }
  }
}

// Notificación del estado de la central, esta función se lanza desde la libería XpressNet ----------------------------------------------------------------------
void notifyXNetPower (uint8_t State){
  switch (State) {
    case csNormal:          // Central operativa
      centralStatus = 0; 
      startup = true;       // Al pasar la central a operativa, activamos el flag para actualizar la posición de los accesorios
      LED_CENTRAL_ON; 
      break;  
    case csEmergencyStop: centralStatus = 5; break; 
    case csShortCircuit: centralStatus = 1; break;       
    case csTrackVoltageOff: centralStatus = 5; break;
    case csServiceMode: centralStatus = 5; break;
  }
  centralStatus = State;
}

#endif
// Fin de la funciones compiladas en función de las opciones de debug ===========================================================================================

//---------------------------------------------------------------------------------------------------------------------------------------------------------------
// BUCLE PRINCIPAL
//---------------------------------------------------------------------------------------------------------------------------------------------------------------
void loop() {
  #if DEBUG_SERIE == 0
    XpressNet.receive();                   // Recibimos datos del bus XpressNet
    if(centralStatus != 0 && (millis() - prevMillis > LEDFLASH * centralStatus))LED_CENTRAL_ONOFF  // Si la central no está disponible hacemos parpadear el LED_CENTRAL   
  #endif 
  if (startup == true && centralStatus == 0) {     // Al startup esperamos que la central esté dispobible, luego consultamos estado de los accesorios 
    #if DEBUG_SERIE == 0 
      AskAccPos();                           // consultamos estado de los accesorios
    #endif 
    startup = false;
  } 
  // Gestión de teclas/pulsaciones 
  if (interruptON == true & TCOblocked == false) {         // Si se ha pulsado una tecla y no hay bloqueo del TCO 
    keyInArray = KeyPressed();             // Calculamos # tecla
    if(keyInArray==1){  //(pgm_read_byte(&PortManaged[keyInArray][0]) == STOP_KEY){  // Si la tecla pulsada corresponde al port con la acción STOP_KEY, controlamos estado de la central
    #if DEBUG_LCD == 1
      DebugLCD("Tecla: ",keyInArray,"",0);
      delay(500);
    #endif
      #if DEBUG_SERIE == 0
        if (XpressNet.getPower() == csNormal) {
          XpressNet.setPower(csTrackVoltageOff);  // Cortamos la alimentación a la vía
        }
        else {
            XpressNet.setPower(csNormal);  //  Pasamos la central a operativa
        }
        #endif
    } 
    if (keyInArray != 0 & centralStatus == 0) {                      // Si tecla valida,procesamos ruta si la central está operativa    
      switch (pgm_read_byte(&PortManaged[keyInArray][0])) {      // determinamos acción en función del tipo de ruta
      case 1:  // Procesamos ruta tipo 1     
        launchRoute_1(keyInArray);
        break;
      case 2:  // Procesamos ruta tipo 2
        launchRoute_2(keyInArray);
        break;
      case 3:  // Procesamos ruta ResetRuta    
        launchReset();
        break;
      case 4:  // procesamos mas de una ruta a la vez de tipos 1 y/o 2
        n = 1;
        while(pgm_read_byte(&PortManaged[keyInArray][n]) != 0){                                    // Procesamos las columnas de la matriz RutasManaged  
          if (pgm_read_byte(&PortManaged[pgm_read_byte(&PortManaged[keyInArray][n])][0]) == 1) {   // Si la ruta es de tipo = 1
            launchRoute_1(pgm_read_byte(&PortManaged[keyInArray][n]));
          }
          else {                                                                                   // Si la ruta es de tipo = 2
            launchRoute_2(pgm_read_byte(&PortManaged[keyInArray][n]));
          }
          n++;
        }
        break; 
      case 5:  // Se pueden definir tantas acciones como se necesiten, configurando el puerto correspondiente y esta estructura "case"
        break;  
      case 6:   
        break;
      case 7:
        break;
      case 8:
        break;
      case 9:        
        break;  
      default:
        break;
        ;
      }          
    }
    // Inicializamos para una nueva lectura
    keyInArray = 0;
    interruptON = false;
    INT_ON; // Se habilita el servicio de interrupción
  }
}



